def exact_div(x, y):
    assert x % y == 0
    return x // y

# hard-coded audio hyperparameters
N_FFT = 400
INPUT_STRIDE=2
N_MELS=80
HOP_LENGTH = 160
CHUNK_LENGTH = 30
SAMPLE_RATE = 16000
MAX_TEXT_TOKEN_LENGTH = 448
N_SAMPLES = CHUNK_LENGTH * SAMPLE_RATE  # 480000 samples in a 30-second chunk
N_FRAMES = exact_div(N_SAMPLES, HOP_LENGTH)  # 3000 frames in a mel spectrogram input

N_SAMPLES_PER_TOKEN = HOP_LENGTH*INPUT_STRIDE  # the initial convolutions has stride 2
FRAMES_PER_SECOND = exact_div(SAMPLE_RATE, HOP_LENGTH)  # 10ms per audio frame
TOKENS_PER_SECOND = exact_div(SAMPLE_RATE, N_SAMPLES_PER_TOKEN)  # 20ms per audio token
TIME_PRECISION = 1/TOKENS_PER_SECOND


import os
import torch
import numpy as np

from . import VADBaseClass
from .. import BASE_PATH


class SegmentVAD(VADBaseClass):
    def __init__(self, 
                 device=None,
                 win_len=0.32,
                 win_step=0.08,
                 batch_size=512,
                 sampling_rate=16000):
        
        super().__init__(sampling_rate=sampling_rate)
        
        if device == None:
            device = 'cuda' if torch.cuda.is_available() else 'cpu'
            
        self.device = device

        if self.device == 'cpu':
            # This is a JIT Scripted model of Nvidia's NeMo Marblenet Model: https://catalog.ngc.nvidia.com/orgs/nvidia/teams/nemo/models/vad_multilingual_marblenet
            self.vad_pp = torch.jit.load(os.path.join(BASE_PATH, "assets/vad_pp_cpu.ts")).to(self.device)
            self.vad_model = torch.jit.load(os.path.join(BASE_PATH, "assets/seg_vad_model_cpu.ts")).to(self.device)
        else:
            self.vad_pp = torch.jit.load(os.path.join(BASE_PATH, "assets/vad_pp_gpu.ts")).to(self.device)
            self.vad_model = torch.jit.load(os.path.join(BASE_PATH, "assets/seg_vad_model_gpu.ts")).to(self.device)
        
        self.vad_pp = torch.jit.load(os.path.join(BASE_PATH, "assets/vad_pp.ts"))
        self.vad_model = torch.jit.load(os.path.join(BASE_PATH, "assets/segment_vad_model.ts"))
        
        self.vad_model.eval()
        self.vad_model.to(self.device)
        
        self.vad_pp.eval()
        self.vad_pp.to(self.device)
        
        self.batch_size = batch_size
        self.win_len = win_len
        self.win_step = win_step
        
        self._init_params()
        
    def _init_params(self):
        self.signal_win_len = int(self.win_len*self.sampling_rate)
        self.signal_win_step = int(self.win_step*self.sampling_rate)
        
    def update_params(self, params={}):
        for key, value in params.items():
            setattr(self, key, value)
        
        self._init_params()
        
    def prepare_input_batch(self, audio_signal):

        num_chunks = (self.signal_win_len//2+len(audio_signal))//self.signal_win_step
        if num_chunks < (self.signal_win_len//2+len(audio_signal))/self.signal_win_step:
            num_chunks += 1

        input_signal = np.zeros((num_chunks, self.signal_win_len), dtype=np.float32)
        input_signal_length = np.zeros(num_chunks, dtype=np.int64)

        chunk_idx = 0
        for idx in range(-1*self.signal_win_len//2, len(audio_signal), self.signal_win_step):
            s_idx = max(idx, 0)
            e_idx = min(idx + self.signal_win_len, len(audio_signal))
            input_signal[chunk_idx][:e_idx-s_idx] = audio_signal[s_idx:e_idx]
            input_signal_length[chunk_idx] = e_idx-s_idx
            chunk_idx += 1

        return input_signal, input_signal_length
    
    @torch.cuda.amp.autocast()
    @torch.no_grad()
    def forward(self, input_signal, input_signal_length):
        x, x_len = self.vad_pp(torch.Tensor(input_signal).to(self.device), 
                               torch.Tensor(input_signal_length).to(self.device))
        logits = self.vad_model(x, x_len)
        logits = torch.softmax(logits, dim=-1)
        return logits[:, 1].detach().cpu().numpy()
    
    def __call__(self, audio_signal):
        
        audio_duration = len(audio_signal)/self.sampling_rate
        
        input_signal, input_signal_length = self.prepare_input_batch(audio_signal)
        
        speech_probs = np.zeros(len(input_signal))
        for s_idx in range(0, len(input_signal), self.batch_size):
            speech_probs[s_idx:s_idx+self.batch_size] = self.forward(input_signal=input_signal[s_idx:s_idx+self.batch_size],
                                                                     input_signal_length=input_signal_length[s_idx:s_idx+self.batch_size])

        vad_times = []
        for idx, prob in enumerate(speech_probs):
            s_time = max(0, (idx-0.5)*self.win_step)
            e_time = min(audio_duration, (idx+0.5)*self.win_step)
            vad_times.append([prob, s_time, e_time])
            
        return np.array(vad_times)



import os
import wave
import tempfile
import numpy as np

import torch
import torch.nn as nn
import torch.nn.functional as F

from multiprocessing.dummy import Pool

from . import BASE_PATH
from .configs import *

silent_file = f"{BASE_PATH}/assets/silent.mp3"

RESAMPLING_ENGINE = 'soxr'
with tempfile.TemporaryDirectory() as tmpdir:
    ret_code = os.system(f'ffmpeg -version')
    if ret_code != 0:
        print(f"Seems 'ffmpeg' is not installed. Please install ffmpeg before using this package!")
    else:
        ret_code = os.system(f'ffmpeg -hide_banner -loglevel panic -i {silent_file} -threads 1 -acodec pcm_s16le -ac 1 -af aresample=resampler={RESAMPLING_ENGINE} -ar 1600 {tmpdir}/tmp.wav -y')

        if ret_code != 0:
            print(f"'ffmpeg' is not built with soxr resampler, using 'swr' resampler. This may degrade performance.")
            RESAMPLING_ENGINE = 'swr'

def load_audio(input_file, sr=16000, return_duration=False):
    
    try:
        with wave.open(input_file, 'rb') as wf:
            if (wf.getframerate() != sr) or (wf.getnchannels() != 1):
                raise Exception("Not a 16kHz wav mono channel file!")
                
            frames = wf.getnframes()
            x = wf.readframes(int(frames))
    except:
        with tempfile.TemporaryDirectory() as tmpdir:
            wav_file = f"{tmpdir}/tmp.wav"
            ret_code = os.system(f'ffmpeg -hide_banner -loglevel panic -i {input_file} -threads 1 -acodec pcm_s16le -ac 1 -af aresample=resampler={RESAMPLING_ENGINE} -ar {sr} {wav_file} -y')
            if ret_code != 0: raise RuntimeError("ffmpeg failed to resample the input audio file, make sure ffmpeg is compiled properly!")
        
            with wave.open(wav_file, 'rb') as wf:
                frames = wf.getnframes()
                x = wf.readframes(int(frames))
    
    audio_signal = np.frombuffer(x, np.int16).flatten().astype(np.float32)/32768.0
    audio_duration = len(audio_signal)/sr
    
    if return_duration:
        return audio_signal, audio_duration
    else:
        return audio_signal


THREAD_POOL_AUDIO_LOADER = Pool(2)
def audio_batch_generator(audio_files):
    return THREAD_POOL_AUDIO_LOADER.imap(load_audio, audio_files)


def pad_or_trim(array, length: int = N_SAMPLES, *, axis: int = -1):
    """
    Pad or trim the audio array to N_SAMPLES, as expected by the encoder.
    """
    
    if torch.is_tensor(array):
        if array.shape[axis] > length:
            array = array.index_select(
                dim=axis, index=torch.arange(length, device=array.device)
            )

        if array.shape[axis] < length:
            pad_widths = [(0, 0)] * array.ndim
            pad_widths[axis] = (0, length - array.shape[axis])
            array = F.pad(array, [pad for sizes in pad_widths[::-1] for pad in sizes])
    else:
        if array.shape[axis] > length:
            array = array.take(indices=range(length), axis=axis)

        if array.shape[axis] < length:
            pad_widths = [(0, 0)] * array.ndim
            pad_widths[axis] = (0, length - array.shape[axis])
            array = np.pad(array, pad_widths)
    
    return array


class TorchSTFT(nn.Module):
    def __init__(self, n_fft, hop_length):
        super().__init__()
        
        self.n_fft = n_fft
        self.hop_length = hop_length
        
        window = torch.hann_window(n_fft)
        self.register_buffer("window", window)
        
    def forward(self, x):
        return torch.stft(x, self.n_fft, self.hop_length, window=self.window, return_complex=True)


class LogMelSpectogram(nn.Module):
    def __init__(self, 
                 n_mels=N_MELS,
                 n_fft=N_FFT,
                 hop_length=HOP_LENGTH,
                 padding=0):
        
        super().__init__()
        
        self.n_fft = n_fft
        self.n_mels = n_mels
        self.hop_length = hop_length
        self.padding = padding
        
        mel_filters = np.load(os.path.join(BASE_PATH, "assets/mel_filters.npz"))
        mel_filters = torch.from_numpy(mel_filters[f"mel_{n_mels}"])
        self.register_buffer("mel_filters", mel_filters)
        
        self.stft = TorchSTFT(n_fft, hop_length)
        
    def get_seq_len(self, seq_len):
        seq_len = torch.floor(seq_len/self.hop_length)
        return seq_len.to(dtype=torch.long)
    
    @torch.no_grad()
    def forward(self, x, seq_len):
        
        seq_len = self.get_seq_len(seq_len.float())
        
        if self.padding > 0:
            x = F.pad(x, (0, self.padding))
            
        x = self.stft(x)
        
        x = x[..., :-1].abs()**2
        x = self.mel_filters@x # mels

        x = torch.clamp(x, min=1e-10).log10() # log_mels
        x = torch.maximum(x, torch.amax(x, dim=(1, 2), keepdims=True) - 8.0) # clip
        x = (x + 4.0) / 4.0 # scale
        
        return x, seq_len


import whisper
from whisper.decoding import DecodingOptions

from .. import WhisperModel
from ...configs import *


ASR_OPTIONS = {
    "beam_size": 1,
    "without_timestamps": True,
    "return_scores": True,
    "return_no_speech_prob": True,
    "patience": 1,
    "length_penalty": 1,
}


class WhisperModelOAI(WhisperModel):
    def __init__(self,
                 model_name: str,
                 device="cuda",
                 compute_type="float16",
                 max_text_token_len=MAX_TEXT_TOKEN_LENGTH,
                 asr_options={},
                 **model_kwargs):

        self.model_name = model_name
        self.asr_options = ASR_OPTIONS
        self.asr_options.update(asr_options)

        self.model = whisper.load_model(model_name)
        self.model.to(device).eval()
        
        self.decode_options = {
            "sample_len": max_text_token_len,
            'fp16': True if compute_type == "float16" else False
        }

        for k, v in self.asr_options.items():
            if hasattr(DecodingOptions, k):
                self.decode_options[k] = v

        super().__init__(
            device=device,
            compute_type=compute_type,
            max_text_token_len=max_text_token_len,
            **model_kwargs
        )

    def update_decode_options(self, params={}):
        self.decode_options.update(params)

        if 'sample_len' in params:
            self.update_params(params={'max_text_token_len': params['sample_len']})
    
    def generate_segment_batched(self, features, prompts, seq_lens, seg_metadata):

        if self.compute_type == "float16":
            features = features.to(self.device).half()

        lang_and_task_pairs = {}
        for _i, _p in enumerate(prompts):
            try:
                lang_and_task_pairs[(_p[-3], _p[-2])].append(_i)
            except:
                lang_and_task_pairs[(_p[-3], _p[-2])] = [_i]


        response = [{} for _ in prompts]
        for (task, lang), idx_list in lang_and_task_pairs.items():
            
            results = self.model.decode(features[idx_list].to(self.device), DecodingOptions(task=task, language=lang, **self.decode_options))

            for idx, result in zip(idx_list, results):
                response[idx]['text'] = result.text.strip()

                if self.asr_options['return_scores']:
                    response[idx]['avg_logprob'] = result.avg_logprob

                if self.asr_options['return_no_speech_prob']:
                    response[idx]['no_speech_prob'] = result.no_speech_prob

        return response


import torch

import numpy as np
from tqdm import tqdm

from .configs import *
from .audio import pad_or_trim, audio_batch_generator, load_audio


def stitch_speech_segments(start_ends, max_len=27.0, max_silent_region=None):

    speech_duration = [end - start for start, end in start_ends]
    
    stitched_speech_segments = []
    
    curr_seg = [0]
    curr_dur = speech_duration[0]
    idx = 1
    while idx < len(start_ends):
        if curr_dur + speech_duration[idx] > max_len:
            stitched_speech_segments.append([start_ends[_] for _ in curr_seg])
            curr_seg = [idx]
            curr_dur = speech_duration[idx]
        else:
            curr_dur += speech_duration[idx]
            curr_seg.append(idx)
            
        idx += 1
        
    stitched_speech_segments.append([start_ends[_] for _ in curr_seg])
    
    if max_silent_region is None:
        return stitched_speech_segments
    
    stitched_speech_segments_joined = []
    for segs in stitched_speech_segments:
        _segs = []
        curr_seg_start_time, curr_seg_end_time = segs[0]
        for i in range(1, len(segs)):
            if (segs[i][0] - curr_seg_end_time) >= max_silent_region:
                _segs.append((curr_seg_start_time, curr_seg_end_time))
                curr_seg_start_time = segs[i][0]

            curr_seg_end_time = segs[i][1]

        _segs.append((curr_seg_start_time, curr_seg_end_time))
        
        stitched_speech_segments_joined.append(_segs)
        
    
    return stitched_speech_segments_joined


class WhisperDataset(torch.utils.data.Dataset):
    def __init__(self, audio_files, lang_codes, tasks, initial_prompts, tokenizer, max_initial_prompt_len, 
                 device="cuda", 
                 dta_padding=48000,
                 without_timestamps=True,
                 use_dynamic_time_axis=False):
        
        self.audio_files = audio_files
        self.lang_codes = lang_codes
        self.tasks = tasks
        self.initial_prompts = initial_prompts
        self.tokenizer = tokenizer
        self.device = device
        self.dta_padding = dta_padding
        self.without_timestamps = without_timestamps
        self.use_dynamic_time_axis = use_dynamic_time_axis
        self.max_initial_prompt_len = max_initial_prompt_len
        
        if type(audio_files[0]) == str:
            self.get_audio_signal = self._get_audio_signal_from_file
        else:
            self.get_audio_signal = self._get_audio_signal_from_array
        
    def _get_audio_signal_from_array(self, item):
        return self.audio_files[item]
        
    def _get_audio_signal_from_file(self, item):
        return load_audio(self.audio_files[item])

    def __len__(self):
        return len(self.audio_files)

    def __getitem__(self, item):
        audio = self.get_audio_signal(item)
        seq_len = audio.shape[-1]
        
        if self.initial_prompts[item]:
            initial_prompt = " " + self.initial_prompts[item].strip()
            initial_prompt_tokens = self.tokenizer.encode(initial_prompt)[-self.max_initial_prompt_len:]
        else:
            initial_prompt_tokens = []
        
        prompt = self.tokenizer.sot_sequence(task=self.tasks[item], lang=self.lang_codes[item])
        
        if self.without_timestamps:
            prompt = prompt + [self.tokenizer.no_timestamps]
            
        return audio, prompt, initial_prompt_tokens, seq_len


class WhisperDataLoader:
    def __init__(self, device, tokenizer, speech_segmenter, 
                 dta_padding=3.0, 
                 without_timestamps=True, 
                 max_speech_len=29.0, 
                 max_initial_prompt_len=223,
                 merge_chunks=True,
                 use_dynamic_time_axis=False):
        
        self.device = device
        self.tokenizer = tokenizer
        self.speech_segmenter = speech_segmenter
        self.dta_padding = int(dta_padding*SAMPLE_RATE)
        self.without_timestamps = without_timestamps
        self.max_speech_len = max_speech_len
        self.max_initial_prompt_len = max_initial_prompt_len
        self.use_dynamic_time_axis = use_dynamic_time_axis
        self.merge_chunks = merge_chunks
        
    def data_collate_fn(self, batch):
        if self.use_dynamic_time_axis:
            max_len = min(max([_[3] for _ in batch]) + self.dta_padding, N_SAMPLES)
        else:
            max_len = N_SAMPLES

        signal_batch = torch.stack([torch.from_numpy(pad_or_trim(_[0], length=max_len)).to(self.device) for _ in batch])
        seq_len = torch.tensor([_[3] for _ in batch]).to(self.device)

        prompt_batch = []
        initial_prompt_max_len = max([len(_[2]) for _ in batch])
        if initial_prompt_max_len:
            for _ in batch: prompt_batch.append([self.tokenizer.sot_prev] + (initial_prompt_max_len-len(_[2]))*[self.tokenizer.silent_token] + _[2] + _[1])
        else:
            for _ in batch: prompt_batch.append(_[1])

        if len(batch[0]) == 5:
            seg_metadata = [_[4] for _ in batch]
            return signal_batch, prompt_batch, seq_len, seg_metadata
        else:
            return signal_batch, prompt_batch, seq_len
    
    def get_segmented_audio_signal(self, audio_signal, file_id, lang, task, initial_prompt, sr=16000):
        start_ends, audio_signal = self.speech_segmenter(audio_signal=audio_signal)

        if initial_prompt:
            initial_prompt = " " + initial_prompt.strip()
            initial_prompt_tokens = self.tokenizer.encode(initial_prompt)[-self.max_initial_prompt_len:]
        else:
            initial_prompt_tokens = []

        prompt = self.tokenizer.sot_sequence(task=task, lang=lang)
        
        if self.without_timestamps:
            prompt.append(self.tokenizer.no_timestamps)
        else:
            prompt.append(self.tokenizer.timestamp_begin)

        segmented_audio_signal = []

        if self.merge_chunks:
            stitched_speech_segments = stitch_speech_segments(start_ends, max_len=self.max_speech_len)
            for stitched_seg in stitched_speech_segments:
                audio = []
                for st, et in stitched_seg:
                    audio.append(audio_signal[int(st*sr):int(et*sr)])

                audio = np.concatenate(audio)
                seq_len = audio.shape[-1]
                seg_metadata = {
                    'file_id': file_id, 
                    'start_time': stitched_seg[0][0], 
                    'end_time': stitched_seg[-1][1], 
                    'stitched_seg': stitched_seg,
                    'lang_code': lang
                }
                segmented_audio_signal.append((audio, prompt, initial_prompt_tokens, seq_len, seg_metadata))
        else:
            for st, et in start_ends:
                audio = audio_signal[int(st*sr):int(et*sr)]
                seq_len = audio.shape[-1]
                segmented_audio_signal.append((audio, prompt, initial_prompt_tokens, seq_len, {'file_id': file_id, 'start_time': st, 'end_time': et}))

        return segmented_audio_signal
    
    def get_data_loader_with_vad(self, audio_files, lang_codes, tasks, initial_prompts, batch_size=16):

        segmented_audio_signal = []
        pbar_update_len = {}
        for file_id, (audio_signal, lang, task, initial_prompt) in enumerate(zip(audio_batch_generator(audio_files), lang_codes, tasks, initial_prompts)):
            new_segmented_audio_signal = self.get_segmented_audio_signal(audio_signal, file_id, lang, task, initial_prompt)
            pbar_update_len[file_id] = 1/len(new_segmented_audio_signal)
            
            segmented_audio_signal = segmented_audio_signal + new_segmented_audio_signal

            while len(segmented_audio_signal) > batch_size:
                batch = segmented_audio_signal[:batch_size]
                segmented_audio_signal = segmented_audio_signal[batch_size:]
                
                signal_batch, prompt_batch, seq_len, seg_metadata = self.data_collate_fn(batch)
                pbar_update = int(sum([pbar_update_len[_['file_id']] for _ in seg_metadata])*100)
                
                yield signal_batch, prompt_batch, seq_len, seg_metadata, pbar_update
        
        signal_batch, prompt_batch, seq_len, seg_metadata = self.data_collate_fn(segmented_audio_signal)
        pbar_update = int(sum([pbar_update_len[_['file_id']] for _ in seg_metadata])*100)

        yield signal_batch, prompt_batch, seq_len, seg_metadata, pbar_update
    
    def get_data_loader(self, audio_files, lang_codes, tasks, initial_prompts, batch_size=16):
        
        dataset = WhisperDataset(audio_files, lang_codes, tasks, initial_prompts, self.tokenizer, 
                                 without_timestamps=self.without_timestamps,
                                 max_initial_prompt_len=self.max_initial_prompt_len,
                                 use_dynamic_time_axis=self.use_dynamic_time_axis)
        
        data_loader = torch.utils.data.DataLoader(dataset, batch_size=batch_size, collate_fn=self.data_collate_fn)
            
        return tqdm(data_loader, desc=f"Transcribing")
    
    def __call__(self, audio_files, lang_codes, tasks, initial_prompts, batch_size=16, use_vad=True):
        if use_vad:
            return self.get_data_loader_with_vad(audio_files, lang_codes, tasks, initial_prompts, batch_size=batch_size)
        else:
            return self.get_data_loader(audio_files, lang_codes, tasks, initial_prompts, batch_size=batch_size)


import os
import torch
import numpy as np

from . import VADBaseClass
from .. import BASE_PATH


class FrameVAD(VADBaseClass):
    def __init__(self, 
                 device=None,
                 chunk_size=15.0,
                 margin_size=1.0,
                 frame_size=0.02,
                 batch_size=4,
                 sampling_rate=16000):
        
        super().__init__(sampling_rate=sampling_rate)
        
        if device == None:
            device = 'cuda' if torch.cuda.is_available() else 'cpu'
            
        self.device = device
        
        if self.device == 'cpu':
            # This is a JIT Scripted model of Nvidia's NeMo Framewise Marblenet Model: https://catalog.ngc.nvidia.com/orgs/nvidia/teams/nemo/models/vad_multilingual_frame_marblenet
            self.vad_pp = torch.jit.load(os.path.join(BASE_PATH, "assets/vad_pp_cpu.ts")).to(self.device)
            self.vad_model = torch.jit.load(os.path.join(BASE_PATH, "assets/frame_vad_model_cpu.ts")).to(self.device)
        else:
            self.vad_pp = torch.jit.load(os.path.join(BASE_PATH, "assets/vad_pp_gpu.ts")).to(self.device)
            self.vad_model = torch.jit.load(os.path.join(BASE_PATH, "assets/frame_vad_model_gpu.ts")).to(self.device)

        self.vad_pp.eval()
        self.vad_model.eval()
        
        self.batch_size = batch_size
        self.frame_size = frame_size
        self.chunk_size = chunk_size
        self.margin_size = margin_size
        
        self._init_params()
        
    def _init_params(self):
        self.signal_chunk_len = int(self.chunk_size*self.sampling_rate)
        self.signal_stride = int(self.signal_chunk_len-2*int(self.margin_size*self.sampling_rate))
        
        self.margin_logit_len = int(self.margin_size/self.frame_size)
        self.signal_to_logit_len = int(self.frame_size*self.sampling_rate)
        
        self.vad_pp.to(self.device)
        self.vad_model.to(self.device)
        
    def update_params(self, params={}):
        for key, value in params.items():
            setattr(self, key, value)
        
        self._init_params()
    
    def prepare_input_batch(self, audio_signal):
        input_signal = []
        input_signal_length = []
        for s_idx in range(0, len(audio_signal), self.signal_stride):
            _signal = audio_signal[s_idx:s_idx+self.signal_chunk_len]
            _signal_len = len(_signal)
            input_signal.append(_signal)
            input_signal_length.append(_signal_len)

            if _signal_len < self.signal_chunk_len:
                input_signal[-1] = np.pad(input_signal[-1], (0, self.signal_chunk_len-_signal_len))
                break

        return input_signal, input_signal_length

    @torch.cuda.amp.autocast()
    @torch.no_grad()
    def forward(self, input_signal, input_signal_length):
        
        all_logits = []
        for s_idx in range(0, len(input_signal), self.batch_size):
            input_signal_pt = torch.stack([torch.tensor(_, device=self.device) for _ in input_signal[s_idx:s_idx+self.batch_size]])
            input_signal_length_pt = torch.tensor(input_signal_length[s_idx:s_idx+self.batch_size], device=self.device)
            
            x, x_len = self.vad_pp(input_signal_pt, input_signal_length_pt)
            logits = self.vad_model(x, x_len)

            for _logits, _len in zip(logits, input_signal_length_pt):
                all_logits.append(_logits[:int(_len/self.signal_to_logit_len)])
        
        if len(all_logits) > 1 and self.margin_logit_len > 0:
            all_logits[0] = all_logits[0][:-self.margin_logit_len]
            all_logits[-1] = all_logits[-1][self.margin_logit_len:]

            for i in range(1, len(all_logits)-1):
                all_logits[i] = all_logits[i][self.margin_logit_len:-self.margin_logit_len]

        all_logits = torch.concatenate(all_logits)
        all_logits = torch.softmax(all_logits, dim=-1)
        
        return all_logits[:, 1].detach().cpu().numpy()
    
    def __call__(self, audio_signal):
        audio_duration = len(audio_signal)/self.sampling_rate
        
        input_signal, input_signal_length = self.prepare_input_batch(audio_signal)
        speech_probs = self.forward(input_signal, input_signal_length)

        vad_times = []
        for idx, prob in enumerate(speech_probs):
            s_time = idx*self.frame_size
            e_time = min(audio_duration, (idx+1)*self.frame_size)
            
            if s_time >= e_time: break
                
            vad_times.append([prob, s_time, e_time])

        return np.array(vad_times)



import os
from platformdirs import user_cache_dir


BASE_PATH = os.path.dirname(__file__)

CACHE_DIR = user_cache_dir("whisper_s2t")
os.makedirs(CACHE_DIR, exist_ok=True)


def load_model(model_identifier="large-v2", 
               backend='CTranslate2', 
               **model_kwargs):
    
    if model_identifier in ['large-v3']:
        model_kwargs['n_mels'] = 128
    elif (model_identifier in ['distil-large-v2']) and (backend.lower() not in ["huggingface", "hf"]):
        print(f"Switching backend to HuggingFace. Distill whisper is only supported with HuggingFace backend.")
        backend = "huggingface"

        model_kwargs['max_speech_len'] = 15.0
        model_kwargs['max_text_token_len'] = 128
    
    if backend.lower() in ["ctranslate2", "ct2"]:
        from .backends.ctranslate2.model import WhisperModelCT2 as WhisperModel

    elif backend.lower() in ["huggingface", "hf"]:
        from .backends.huggingface.model import WhisperModelHF as WhisperModel

        if 'distil' in model_identifier:
            model_identifier = f"distil-whisper/{model_identifier}"
        else:
            model_identifier = f"openai/whisper-{model_identifier}"

    elif backend.lower() in ["openai", "oai"]:
        from .backends.openai.model import WhisperModelOAI as WhisperModel

    else:
        raise ValueError(f"Backend name '{backend}' is invalid. Only following options are available: ['CTranslate2', 'HuggingFace', 'OpenAI']")
        
    return WhisperModel(model_identifier, **model_kwargs)
        


import os
import tokenizers
import ctranslate2
import numpy as np

from .tokenizer import Tokenizer
from .hf_utils import download_model


from .. import WhisperModel
from ...configs import *


FAST_ASR_OPTIONS = {
    "beam_size": 1,
    "best_of": 1, # Placeholder
    "patience": 1,
    "length_penalty": 1,
    "repetition_penalty": 1.01,
    "no_repeat_ngram_size": 0,
    "compression_ratio_threshold": 2.4, # Placeholder
    "log_prob_threshold": -1.0, # Placeholder
    "no_speech_threshold": 0.5, # Placeholder
    "prefix": None, # Placeholder
    "suppress_blank": True,
    "suppress_tokens": [-1],
    "without_timestamps": True,
    "max_initial_timestamp": 1.0,
    "word_timestamps": False, # Placeholder
    "sampling_temperature": 1.0,
    "return_scores": True,
    "return_no_speech_prob": True,
    "word_aligner_model": 'tiny',
}


BEST_ASR_CONFIG = {
    "beam_size": 5,
    "best_of": 1, # Placeholder
    "patience": 2,
    "length_penalty": 1,
    "repetition_penalty": 1.01,
    "no_repeat_ngram_size": 0,
    "compression_ratio_threshold": 2.4, # Placeholder
    "log_prob_threshold": -1.0, # Placeholder
    "no_speech_threshold": 0.5, # Placeholder
    "prefix": None, # Placeholder
    "suppress_blank": True,
    "suppress_tokens": [-1],
    "without_timestamps": True,
    "max_initial_timestamp": 1.0,
    "word_timestamps": False, # Placeholder
    "sampling_temperature": 1.0,
    "return_scores": True,
    "return_no_speech_prob": True,
    "word_aligner_model": 'tiny',
}


class WhisperModelCT2(WhisperModel):
    def __init__(self,
                 model_name_or_path: str,
                 cpu_threads=4,
                 num_workers=1,
                 device="cuda",
                 device_index=0,
                 compute_type="float16",
                 max_text_token_len=MAX_TEXT_TOKEN_LENGTH,
                 asr_options={},
                 **model_kwargs):

        
        # Get local model path or download from huggingface
        if os.path.isdir(model_name_or_path):
            self.model_path = model_name_or_path
        else:
            self.model_path = download_model(model_name_or_path)
        
        # Load model
        self.model = ctranslate2.models.Whisper(self.model_path,
                                                device=device,
                                                device_index=device_index,
                                                compute_type=compute_type,
                                                intra_threads=cpu_threads,
                                                inter_threads=num_workers)
        
        # Load tokenizer
        tokenizer_file = os.path.join(self.model_path, "tokenizer.json")
        tokenizer = Tokenizer(tokenizers.Tokenizer.from_file(tokenizer_file), self.model.is_multilingual)

        # ASR Options
        self.asr_options = FAST_ASR_OPTIONS
        self.asr_options.update(asr_options)

        if self.asr_options['word_timestamps']:
            self.aligner_model_path = download_model(self.asr_options['word_aligner_model'])
            self.aligner_model = ctranslate2.models.Whisper(self.aligner_model_path,
                                                            device=device,
                                                            device_index=device_index,
                                                            compute_type=compute_type,
                                                            intra_threads=cpu_threads,
                                                            inter_threads=num_workers)
        
        self.generate_kwargs = {
            "max_length": max_text_token_len,
            "return_scores": self.asr_options['return_scores'],
            "return_no_speech_prob": self.asr_options['return_no_speech_prob'],
            "length_penalty": self.asr_options['length_penalty'],
            "repetition_penalty": self.asr_options['repetition_penalty'],
            "no_repeat_ngram_size": self.asr_options['no_repeat_ngram_size'],
            "beam_size": self.asr_options['beam_size'],
            "patience": self.asr_options['patience'],
            "suppress_blank": self.asr_options['suppress_blank'],
            "suppress_tokens": self.asr_options['suppress_tokens'],
            "max_initial_timestamp_index": int(round(self.asr_options['max_initial_timestamp']/TIME_PRECISION)),
            "sampling_temperature": self.asr_options['sampling_temperature'],
        }

        super().__init__(
            tokenizer=tokenizer,
            device=device,
            device_index=device_index,
            compute_type=compute_type,
            max_text_token_len=max_text_token_len,
            **model_kwargs
        )

    def update_generation_kwargs(self, params={}):
        self.generate_kwargs.update(params)

        if 'max_text_token_len' in params:
            self.update_params(params={'max_text_token_len': params['max_text_token_len']})
    
    def encode(self, features):
        """
        [Not Used]
        """
        
        features = ctranslate2.StorageView.from_array(features.contiguous())
        return self.model.encode(features)

    def assign_word_timings(self, alignments, text_token_probs, words, word_tokens):
        text_indices = np.array([pair[0] for pair in alignments])
        time_indices = np.array([pair[1] for pair in alignments])
    
        if len(word_tokens) <= 1:
            return []
            
        word_boundaries = np.pad(np.cumsum([len(t) for t in word_tokens[:-1]]), (1, 0))
        if len(word_boundaries) <= 1:
            return []
    
        jumps = np.pad(np.diff(text_indices), (1, 0), constant_values=1).astype(bool)
        jump_times = time_indices[jumps]*TIME_PRECISION
        start_times = jump_times[word_boundaries[:-1]]
        end_times = jump_times[word_boundaries[1:]]
        word_probs = [
            np.mean(text_token_probs[i:j])
            for i, j in zip(word_boundaries[:-1], word_boundaries[1:])
        ]
    
        return [
            dict(
                word=word, start=round(start, 2), end=round(end, 2), prob=round(prob, 2)
            )
            for word, start, end, prob in zip(
                words, start_times, end_times, word_probs
            )
        ]

    def align_words(self, features, texts, text_tokens, sot_seqs, seq_lens, seg_metadata):
        lang_codes = [_['lang_code'] for _ in seg_metadata]
        word_tokens = self.tokenizer.split_to_word_tokens_batch(texts, text_tokens, lang_codes)

        start_seq_wise_req = {}
        for _idx, _sot_seq in enumerate(sot_seqs):
            try:
                # print(_sot_seq)
                start_seq_wise_req[_sot_seq].append(_idx)
            except:
                start_seq_wise_req[_sot_seq] = [_idx]

        token_alignments = [[] for _ in seg_metadata]
        for start_seq, req_idx in start_seq_wise_req.items():
            res = self.aligner_model.align(ctranslate2.StorageView.from_array(features[req_idx]), 
                                           start_sequence=list(start_seq), 
                                           text_tokens=[text_tokens[_] for _ in req_idx],
                                           num_frames=list(seq_lens[req_idx].detach().cpu().numpy()), 
                                           median_filter_width=7)

            for _res, _req_idx in zip(res, req_idx):
                token_alignments[_req_idx] = _res

        word_timings = []
        for _idx, _seg_metadata in enumerate(seg_metadata):
            _word_timings = self.assign_word_timings(token_alignments[_idx].alignments, 
                                                     token_alignments[_idx].text_token_probs, 
                                                     word_tokens[_idx][0], 
                                                     word_tokens[_idx][1])
        
            stitched_seg = _seg_metadata['stitched_seg']

            current_seg_idx = 0
            current_offset = _seg_metadata['start_time']
        
            for w in _word_timings:
                while (w['start'] + current_offset) >= stitched_seg[current_seg_idx][1]:
                    current_seg_idx += 1
                    current_offset += (stitched_seg[current_seg_idx][0]-stitched_seg[current_seg_idx-1][1])
        
                w['start'] += current_offset
                w['end'] += current_offset
        
            word_timings.append(_word_timings)

        return word_timings
    
    def generate_segment_batched(self, features, prompts, seq_lens, seg_metadata):
        
        if self.device == 'cpu':
            features = np.ascontiguousarray(features.detach().numpy())
        else:
            features = features.contiguous()

        result = self.model.generate(ctranslate2.StorageView.from_array(features),
                                     prompts,
                                     **self.generate_kwargs)
        
        texts = self.tokenizer.decode_batch([x.sequences_ids[0] for x in result])
        
        response = []
        for idx, r in enumerate(result):
            response.append({'text': texts[idx].strip()})

            if self.generate_kwargs['return_scores']:
                seq_len = len(r.sequences_ids[0])
                cum_logprob = r.scores[0]*(seq_len**self.generate_kwargs['length_penalty'])
                response[-1]['avg_logprob'] = cum_logprob/(seq_len + 1)

            if self.generate_kwargs['return_no_speech_prob']:
                response[-1]['no_speech_prob'] = r.no_speech_prob

        if self.asr_options['word_timestamps']:
            text_tokens = [x.sequences_ids[0]+[self.tokenizer.eot] for x in result]
            sot_seqs = [tuple(_[-4:]) for _ in prompts]
            word_timings = self.align_words(features, texts, text_tokens, sot_seqs, seq_lens, seg_metadata)

            for _response, _word_timings in zip(response, word_timings):
                _response['word_timestamps'] = _word_timings

        return response


# https://github.com/guillaumekln/faster-whisper/blob/master/faster_whisper/utils.py

import os
import re
import requests

import huggingface_hub
from typing import List, Optional

from ... import CACHE_DIR


os.makedirs(f"{CACHE_DIR}/models", exist_ok=True)


_MODELS = {
    "tiny.en": "Systran/faster-whisper-tiny.en",
    "tiny": "Systran/faster-whisper-tiny",
    "base.en": "Systran/faster-whisper-base.en",
    "base": "Systran/faster-whisper-base",
    "small.en": "Systran/faster-whisper-small.en",
    "small": "Systran/faster-whisper-small",
    "medium.en": "Systran/faster-whisper-medium.en",
    "medium": "Systran/faster-whisper-medium",
    "large-v1": "Systran/faster-whisper-large-v1",
    "large-v2": "Systran/faster-whisper-large-v2",
    "large-v3": "Systran/faster-whisper-large-v3",
    "large": "Systran/faster-whisper-large-v3",
}


def available_models() -> List[str]:
    """Returns the names of available models."""
    return list(_MODELS.keys())


def download_model(
    size_or_id: str,
    output_dir: Optional[str] = None,
    local_files_only: bool = False,
    cache_dir: Optional[str] = None,
):
    """Downloads a CTranslate2 Whisper model from the Hugging Face Hub.

    Args:
      size_or_id: Size of the model to download from https://huggingface.co/guillaumekln
        (tiny, tiny.en, base, base.en, small, small.en medium, medium.en, large-v1, large-v2,
        large), or a CTranslate2-converted model ID from the Hugging Face Hub
        (e.g. guillaumekln/faster-whisper-large-v2).
      output_dir: Directory where the model should be saved. If not set, the model is saved in
        the cache directory.
      local_files_only:  If True, avoid downloading the file and return the path to the local
        cached file if it exists.
      cache_dir: Path to the folder where cached files are stored.

    Returns:
      The path to the downloaded model.

    Raises:
      ValueError: if the model size is invalid.
    """
    if re.match(r".*/.*", size_or_id):
        repo_id = size_or_id
    else:
        repo_id = _MODELS.get(size_or_id)
        if repo_id is None:
            raise ValueError(
                "Invalid model size '%s', expected one of: %s"
                % (size_or_id, ", ".join(_MODELS.keys()))
            )

    allow_patterns = [
        "config.json",
        "model.bin",
        "tokenizer.json",
        "vocabulary.*",
    ]

    kwargs = {
        "local_files_only": local_files_only,
        "allow_patterns": allow_patterns,
    }

    if output_dir is not None:
        kwargs["local_dir"] = output_dir
        kwargs["local_dir_use_symlinks"] = False

    if cache_dir is not None:
        kwargs["cache_dir"] = cache_dir
    else:
        kwargs["cache_dir"] = f"{CACHE_DIR}/models"

    try:
        return huggingface_hub.snapshot_download(repo_id, **kwargs)
    except (
        huggingface_hub.utils.HfHubHTTPError,
        requests.exceptions.ConnectionError,
    ) as exception:
        print(exception)
        logger = get_logger()
        logger.warning(
            "An error occured while synchronizing the model %s from the Hugging Face Hub:\n%s",
            repo_id,
            exception,
        )
        logger.warning(
            "Trying to load the model directly from the local cache, if it exists."
        )

        kwargs["local_files_only"] = True
        return huggingface_hub.snapshot_download(repo_id, **kwargs)



import torch

from transformers import WhisperProcessor, WhisperForConditionalGeneration

from .. import WhisperModel
from ...configs import *


ASR_OPTIONS = {
    "beam_size": 1,
    "without_timestamps": True,
    "return_scores": False,
    "return_no_speech_prob": False,
    "use_flash_attention": True,
    "use_better_transformer": False,
}


COMPUTE_TYPE_TO_TORCH_DTYPE = {
    "float16": torch.float16
}


class WhisperModelHF(WhisperModel):
    def __init__(self,
                 model_name: str,
                 device="cuda",
                 compute_type="float16",
                 max_text_token_len=MAX_TEXT_TOKEN_LENGTH,
                 asr_options={},
                 **model_kwargs):

        self.model_name = model_name
        self.asr_options = ASR_OPTIONS
        self.asr_options.update(asr_options)

        self.processor = WhisperProcessor.from_pretrained(self.model_name)
        self.model = WhisperForConditionalGeneration.from_pretrained(self.model_name, 
                                                                     torch_dtype=COMPUTE_TYPE_TO_TORCH_DTYPE.get(compute_type, torch.float32), 
                                                                     low_cpu_mem_usage=True, 
                                                                     use_safetensors=True,
                                                                     use_flash_attention_2=self.asr_options["use_flash_attention"])
        self.model.config.forced_decoder_ids = None
        self.model.to(device).eval()

        if self.asr_options["use_better_transformer"]:
            self.model = self.model.to_bettertransformer()
        
        self.generate_kwargs = {
            "max_new_tokens": max_text_token_len,
            "num_beams": self.asr_options['beam_size'],
            "return_timestamps": not self.asr_options['without_timestamps'],
        }

        super().__init__(
            device=device,
            compute_type=compute_type,
            max_text_token_len=max_text_token_len,
            **model_kwargs
        )

    def update_generation_kwargs(self, params={}):
        self.generate_kwargs.update(params)

        if 'max_new_tokens' in params:
            self.update_params(params={'max_text_token_len': params['max_new_tokens']})
    
    def generate_segment_batched(self, features, prompts, seq_lens, seg_metadata):
        if self.compute_type == "float16":
            features = features.to(self.device).half()

        lang_and_task_pairs = {}
        for _i, _p in enumerate(prompts):
            try:
                lang_and_task_pairs[(_p[-3], _p[-2])].append(_i)
            except:
                lang_and_task_pairs[(_p[-3], _p[-2])] = [_i]


        response = [{} for _ in prompts]
        for (task, lang), idx_list in lang_and_task_pairs.items():
            predicted_ids = self.model.generate(features[idx_list], 
                                                task=task,
                                                language=lang,
                                                **self.generate_kwargs)
        
            results = self.processor.batch_decode(predicted_ids, skip_special_tokens=True)

            for idx, text in zip(idx_list, results):
                response[idx]['text'] = text.strip()

        return response


import torch
from tqdm import tqdm
from abc import ABC, abstractmethod

from ..configs import *
from ..data import WhisperDataLoader
from ..audio import LogMelSpectogram
from ..speech_segmenter import SpeechSegmenter


class NoneTokenizer:
    def __init__(self):
        self.sot_prev = 0
        self.silent_token = 0
        self.no_timestamps = 0
        self.timestamp_begin = 0
    
    def sot_sequence(self, task=None, lang=None):
        return [task, lang]

    def encode(self, text):
        return [0]


class WhisperModel(ABC):
    def __init__(self,
                 tokenizer=None,
                 vad_model=None,
                 n_mels=80,
                 device="cuda",
                 device_index=0,
                 compute_type="float16",
                 merge_chunks=True,
                 dta_padding=3.0,
                 use_dynamic_time_axis = False,
                 max_speech_len=29.0,
                 max_text_token_len=MAX_TEXT_TOKEN_LENGTH,
                 without_timestamps=True,
                 speech_segmenter_options={}):
        
        # Configure Params
        self.device = device
        self.device_index = device_index
        self.compute_type = compute_type

        self.n_mels = n_mels
        self.merge_chunks = merge_chunks
        self.max_speech_len = max_speech_len

        self.dta_padding = dta_padding
        self.use_dynamic_time_axis = use_dynamic_time_axis

        self.without_timestamps = without_timestamps
        self.max_text_token_len = max_text_token_len

        self.vad_model = vad_model
        self.speech_segmenter_options = speech_segmenter_options
        self.speech_segmenter_options['max_seg_len'] = self.max_speech_len

        # Tokenizer
        if tokenizer is None:
            tokenizer = NoneTokenizer()

        self.tokenizer = tokenizer

        self._init_dependables()


    def _init_dependables(self):
        # Rescaled Params
        self.dta_padding = int(self.dta_padding*SAMPLE_RATE)
        self.max_initial_prompt_len = self.max_text_token_len//2 -1

        # Load Pre Processor
        self.preprocessor = LogMelSpectogram(n_mels=self.n_mels).to(self.device)

        # Load Speech Segmenter
        self.speech_segmenter = SpeechSegmenter(self.vad_model, device=self.device, **self.speech_segmenter_options)

        # Load Data Loader
        self.data_loader = WhisperDataLoader(
            self.device, self.tokenizer, self.speech_segmenter, 
            dta_padding=self.dta_padding,
            without_timestamps=self.without_timestamps, 
            max_speech_len=self.max_speech_len, 
            max_initial_prompt_len=self.max_initial_prompt_len, 
            use_dynamic_time_axis=self.use_dynamic_time_axis,
            merge_chunks=self.merge_chunks
        )

    def update_params(self, params={}):
        for key, value in params.items():
            setattr(self, key, value)
        
        self._init_dependables()

    
    @abstractmethod
    def generate_segment_batched(self, features, prompts):
        pass
        
    @torch.no_grad()
    def transcribe(self, audio_files, lang_codes=None, tasks=None, initial_prompts=None, batch_size=8):
        
        if lang_codes == None:
            lang_codes = len(audio_files)*['en']
            
        if tasks == None:
            tasks = len(audio_files)*['transcribe']
        
        if initial_prompts == None:
            initial_prompts = len(audio_files)*[None]
            
        responses = []
        for signals, prompts, seq_len in self.data_loader(audio_files, lang_codes, tasks, initial_prompts, batch_size=batch_size, use_vad=False):
            mels, seq_len = self.preprocessor(signals, seq_len)
            res = self.generate_segment_batched(mels.to(self.device), prompts)
            responses.extend(res)
        
        return responses
    
    @torch.no_grad()
    def transcribe_with_vad(self, audio_files, lang_codes=None, tasks=None, initial_prompts=None, batch_size=8):
        
        if lang_codes == None:
            lang_codes = len(audio_files)*['en']
            
        if tasks == None:
            tasks = len(audio_files)*['transcribe']
        
        if initial_prompts == None:
            initial_prompts = len(audio_files)*[None]
            
        responses = [[] for _ in audio_files]
        
        pbar_pos = 0
        with tqdm(total=len(audio_files)*100, desc=f"Transcribing") as pbar:
            for signals, prompts, seq_len, seg_metadata, pbar_update in self.data_loader(audio_files, lang_codes, tasks, initial_prompts, batch_size=batch_size):
                mels, seq_len = self.preprocessor(signals, seq_len)
                res = self.generate_segment_batched(mels.to(self.device), prompts, seq_len, seg_metadata)

                for res_idx, _seg_metadata in enumerate(seg_metadata):
                    responses[_seg_metadata['file_id']].append({**res[res_idx],
                                                                'start_time': round(_seg_metadata['start_time'], 3),
                                                                'end_time': round(_seg_metadata['end_time'], 3)})
                
                if (pbar_pos) <= pbar.total:
                    pbar_pos += pbar_update
                    pbar.update(pbar_update)
            
            pbar.update(pbar.total-pbar_pos)
        
        return responses


from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("requirements.txt", "r", encoding="utf-8") as f:
    requirements = f.read().splitlines()

setup(
    name="whisper_s2t",
    version="1.2.0",
    description="An Optimized Speech-to-Text Pipeline for the Whisper Model.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires=">=3.8",
    author="Shashi Kant Gupta",
    url="https://github.com/shashikg/WhisperS2T",
    license="MIT",
    packages=find_packages(exclude=["tests*"]),
    install_requires=requirements,
    package_data={
        '': ['assets/*'],
    },
    include_package_data=True,
)



import numpy as np
from abc import ABC, abstractmethod

from ..audio import load_audio


class VADBaseClass(ABC):
    def __init__(self, sampling_rate=16000):
        self.sampling_rate = sampling_rate
        
    @abstractmethod
    def update_params(self, params={}):
        pass
    
    @abstractmethod
    def __call__(self, audio_signal, batch_size=4):
        pass
    

class SpeechSegmenter:
    def __init__(self, vad_model=None,
                 device=None,
                 frame_size=0.02,
                 min_seg_len=0.08,
                 max_seg_len=29.0,
                 max_silent_region=0.6,
                 padding=0.2,
                 eos_thresh=0.3,
                 bos_thresh=0.3,
                 cut_factor=2,
                 sampling_rate=16000):
        
        if vad_model is None:
            from .frame_vad import FrameVAD
            vad_model = FrameVAD(device=device)
        
        self.vad_model = vad_model
        
        self.sampling_rate = sampling_rate
        self.padding = padding
        self.frame_size = frame_size
        self.min_seg_len = min_seg_len
        self.max_seg_len = max_seg_len
        self.max_silent_region = max_silent_region
        
        self.eos_thresh = eos_thresh
        self.bos_thresh = bos_thresh
        
        self.cut_factor = cut_factor
        self.cut_idx = int(self.max_seg_len/(self.cut_factor*self.frame_size))
        self.max_idx_in_seg = self.cut_factor*self.cut_idx
        
    def update_params(self, params={}):
        for key, value in params.items():
            setattr(self, key, value)
        
        self.cut_idx = int(self.max_seg_len/(self.cut_factor*self.frame_size))
        self.max_idx_in_seg = self.cut_factor*self.cut_idx
        
    def update_vad_model_params(self, params={}):
        self.vad_model.update_params(params=params)
        
    def okay_to_merge(self, speech_probs, last_seg, curr_seg):
        conditions = [
            (speech_probs[curr_seg['start']][1]-speech_probs[last_seg['end']][2]) < self.max_silent_region,
            (speech_probs[curr_seg['end']][2]-speech_probs[last_seg['start']][1]) <= self.max_seg_len, 
        ]

        return all(conditions)
        
    def get_speech_segments(self, speech_probs):
        
        speech_flag, start_idx = False, 0
        speech_segments = []
        for idx, (speech_prob, st, et) in enumerate(speech_probs):
            if speech_flag:
                if speech_prob < self.eos_thresh:
                    speech_flag = False
                    curr_seg = {'start': start_idx, 'end': idx-1}

                    if len(speech_segments) and self.okay_to_merge(speech_probs, speech_segments[-1], curr_seg):
                        speech_segments[-1]['end'] = curr_seg['end']
                    else:
                        speech_segments.append(curr_seg)

            elif speech_prob >= self.bos_thresh:
                speech_flag = True
                start_idx = idx

        if speech_flag:
            curr_seg = {'start': start_idx, 'end': len(speech_probs)-1}

            if len(speech_segments) and self.okay_to_merge(speech_probs, speech_segments[-1], curr_seg):
                speech_segments[-1]['end'] = curr_seg['end']
            else:
                speech_segments.append(curr_seg)
        
        speech_segments = [_ for _ in speech_segments if (speech_probs[_['end']][2]-speech_probs[_['start']][1]) > self.min_seg_len]

        start_ends = []
        for _ in speech_segments:
            first_idx = len(start_ends)
            start_idx, end_idx = _['start'], _['end']
            while (end_idx-start_idx) > self.max_idx_in_seg:
                _start_idx = int(start_idx + self.cut_idx)
                _end_idx = int(min(end_idx, start_idx + self.max_idx_in_seg))

                new_end_idx = _start_idx+np.argmin(speech_probs[_start_idx:_end_idx, 0])
                start_ends.append([speech_probs[start_idx][1], speech_probs[new_end_idx][2]])
                start_idx = new_end_idx+1
            
            start_ends.append([speech_probs[start_idx][1], speech_probs[end_idx][2]+self.padding])
            start_ends[first_idx][0] = start_ends[first_idx][0]-self.padding
        
        return start_ends
             
    
    def __call__(self, input_file=None, audio_signal=None):
        if audio_signal is None:
            audio_signal, audio_duration = load_audio(input_file, sr=self.sampling_rate, return_duration=True)
        else:
            audio_duration = len(audio_signal)/self.sampling_rate
            
        speech_probs = self.vad_model(audio_signal)
        start_ends = self.get_speech_segments(speech_probs)
        
        start_ends[0][0] = max(0.0, start_ends[0][0]) # fix edges
        start_ends[-1][1] = min(audio_duration, start_ends[-1][1]) # fix edges
        
        return start_ends, audio_signal


import os
from functools import cached_property

from ... import BASE_PATH


_TASKS = (
    "transcribe",
    "translate",
)


with open(os.path.join(BASE_PATH, "assets/lang_codes.txt"), 'r') as f:
    _LANGUAGE_CODES = [_ for _ in f.read().split("\n") if _]


class Tokenizer:
    def __init__(self, tokenizer, multilingual):
        
        self.tokenizer = tokenizer
        self.multilingual = multilingual
        
        if self.multilingual:
            self.task_to_token_id = {task: self.tokenizer.token_to_id(f"<|{task}|>") for task in _TASKS}
            self.lang_code_to_token_id = {lang: self.tokenizer.token_to_id(f"<|{lang}|>") for lang in _LANGUAGE_CODES}
        else:
            self.task_to_token_id = None
            self.lang_code_to_token_id = None

    @cached_property
    def transcribe(self) -> int:
        return self.tokenizer.token_to_id("<|transcribe|>")

    @cached_property
    def translate(self) -> int:
        return self.tokenizer.token_to_id("<|translate|>")
    
    @cached_property
    def silent_token(self) -> int:
        return self.encode(" ")[0]

    @cached_property
    def sot(self) -> int:
        return self.tokenizer.token_to_id("<|startoftranscript|>")

    @cached_property
    def sot_lm(self) -> int:
        return self.tokenizer.token_to_id("<|startoflm|>")

    @cached_property
    def sot_prev(self) -> int:
        return self.tokenizer.token_to_id("<|startofprev|>")

    @cached_property
    def eot(self) -> int:
        return self.tokenizer.token_to_id("<|endoftext|>")

    @cached_property
    def no_timestamps(self) -> int:
        return self.tokenizer.token_to_id("<|notimestamps|>")

    @property
    def timestamp_begin(self) -> int:
        return self.no_timestamps + 1

    def sot_sequence(self, task=None, lang=None):
        sequence = [self.sot]
        
        if self.multilingual:
            sequence.append(self.lang_code_to_token_id[lang])
            sequence.append(self.task_to_token_id[task])

        return sequence

    def encode(self, text):
        return self.tokenizer.encode(text, add_special_tokens=False).ids

    def decode(self, tokens):
        text_tokens = [token for token in tokens if token < self.eot]
        return self.tokenizer.decode(text_tokens)
    
    def decode_batch(self, tokens):
        res = []
        for tk in tokens:
            res.append([token for token in tk if token < self.eot])

        return self.tokenizer.decode_batch(res)
    
    def split_tokens_on_unicode(self, text, tokens):
        replacement_char = "\ufffd"
    
        subwords, subword_tokens_list, current_tokens = [], [], []
        unicode_offset, word_finished = 0, False
        
        for token in tokens:
            current_tokens.append(token)
            decoded = self.decode(current_tokens)
    
            try:
                replacement_char_index = decoded.index(replacement_char) + unicode_offset
                if (replacement_char_index < len(text)) and (text[replacement_char_index] == replacement_char):
                    word_finished = True
            except ValueError:
                word_finished = True
    
            if word_finished:
                subwords.append(decoded)
                subword_tokens_list.append(current_tokens)
                
                current_tokens = []
                word_finished = False
                unicode_offset += len(decoded)
    
        return subwords, subword_tokens_list

    def split_tokens_on_spaces(self, text, tokens):
        subwords, subword_tokens_list = self.split_tokens_on_unicode(text, tokens)
        words = []
        word_tokens = []
    
        for subword, subword_tokens in zip(subwords, subword_tokens_list):
            conditions = [
                subword_tokens[0] >= self.eot, # special
                subword.startswith(" "), # with_space
                # subword.strip() in string.punctuation, # punctuation
                len(words) == 0
            ]
            
            if any(conditions):
                words.append(subword.strip())
                word_tokens.append(subword_tokens)
            else:
                words[-1] = words[-1] + subword
                word_tokens[-1].extend(subword_tokens)
    
        return words, word_tokens
    
    def split_to_word_tokens(self, text, tokens, lang_code):
        if lang_code in {"zh", "ja", "th", "lo", "my", "yue"}:
            # These languages don't typically use spaces, so it is difficult to split words
            # without morpheme analysis. Here, we instead split words at any
            # position where the tokens are decoded as valid unicode points
            return self.split_tokens_on_unicode(text, tokens)
    
        return self.split_tokens_on_spaces(text, tokens)

    def split_to_word_tokens_batch(self, text_batch, tokens_batch, lang_code_batch):
        res = []
        for text, tokens, lang_code in zip(text_batch, tokens_batch, lang_code_batch):
            res.append(self.split_to_word_tokens(text, tokens, lang_code))
    
        return res